use chrono::UTC;
use std::fmt;
use std::hash::{Hash, Hasher};
use std::collections::BTreeMap;
use std::collections::HashMap;
use std::collections::hash_map::DefaultHasher;
use serde::{Deserialize, Deserializer};
use regex::Regex;

#[derive(Serialize, Deserialize)]
pub struct Token {
    pub access_token: String,
    pub expires_in: u16,
    pub token_type: String,
    pub scope: String,
    pub session_id: String,
    pub refresh_token: String,
    pub user_id: String,
    #[serde(default = "timestamp")]
    pub timestamp: i64,
}

impl Token {
    pub fn is_expired(&self) -> bool {
        let now = UTC::now().timestamp();

        now > self.timestamp + self.expires_in as i64
    }
}

fn timestamp() -> i64 {
    UTC::now().timestamp()
}

fn default_list() -> Vec<String> {
    Vec::new()
}

#[derive(Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Config {
    #[serde(default)]
    pub game_storage: String,
    #[serde(default)]
    pub movie_storage: String,
    #[serde(default = "default_list")]
    pub os_filters: Vec<String>,
    #[serde(default = "default_list")]
    pub language_filters: Vec<String>,
    #[serde(default = "default_list")]
    pub resolution_filters: Vec<String>,
    #[serde(default)]
    pub skip_movies: bool,
    #[serde(default)]
    pub skip_games: bool,
}

impl Config {
    pub fn new() -> Config {
        Config {
            game_storage: String::from("."),
            movie_storage: String::from("."),
            os_filters: Vec::new(),
            language_filters: Vec::new(),
            resolution_filters: Vec::new(),
            skip_movies: false,
            skip_games: false,
        }
    }
}

#[derive(Hash)]
#[derive(Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Content {
    #[serde(skip_deserializing)]
    pub id: u64,
    #[serde(deserialize_with = "normalize_title")]
    pub title: String,
    #[serde(skip_deserializing)]
    #[serde(rename(deserialize = "cdKey"))]
    pub cd_keys: BTreeMap<String, String>,
    #[serde(skip_deserializing)]
    pub is_movie: bool,
    #[serde(skip_deserializing)]
    pub data: Vec<Data>,
    pub extras: Vec<Extra>,
}

impl fmt::Display for Content {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "({})", self.title)
    }
}

fn normalize_title<D>(deserializer: D) -> Result<String, D::Error>
    where D: Deserializer
{
    let raw_title = String::deserialize(deserializer)?.replace(":", " - ");

    let regex_normalize = Regex::new(r"[^'^\w^\s^-]+").unwrap();
    let regex_whitespace = Regex::new(r"\s\s+").unwrap();

    let title_normalized = regex_normalize
        .replace_all(raw_title.as_str(), "")
        .to_string();

    let title_whitespace = regex_whitespace
        .replace_all(title_normalized.as_str(), " ")
        .to_string();

    Ok(title_whitespace)
}

#[derive(Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct ContentInfo {
    #[serde(skip_deserializing)]
    pub id: u64,
    pub hash: u64,
    pub title: String,
    pub cd_keys: BTreeMap<String, String>,
    pub data: HashMap<String, DataInfo>,
    pub extras: HashMap<String, ExtraInfo>,
}

impl ContentInfo {
    pub fn new() -> ContentInfo {
        ContentInfo {
            id: 0,
            hash: 0,
            title: String::new(),
            cd_keys: BTreeMap::new(),
            data: HashMap::new(),
            extras: HashMap::new(),
        }
    }
}

#[derive(Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct DataInfo {
    pub hash: u64,
    pub filename: String,
    pub language: String,
}

#[derive(Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct ExtraInfo {
    pub hash: u64,
    pub filename: String,
    pub name: String,
}

#[derive(Hash)]
#[derive(Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Data {
    pub manual_url: String,
    pub version: String,
    pub os: String,
    pub language: String,
}

impl fmt::Display for Data {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f,
               "({}, {}, {}, {})",
               self.manual_url,
               self.version,
               self.os,
               self.language)
    }
}

#[derive(Hash)]
#[derive(Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Extra {
    pub manual_url: String,
    pub name: String,
}

impl fmt::Display for Extra {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "({}, {})", self.manual_url, self.name)
    }
}

pub fn get_hash<T>(obj: &T) -> u64
    where T: Hash
{
    let mut s = DefaultHasher::new();
    obj.hash(&mut s);
    s.finish()
}
